/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.wsas.admin.service.logging;

import org.apache.axis2.AxisFault;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.Appender;
import org.apache.log4j.AppenderSkeleton;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.net.SyslogAppender;
import org.wso2.wsas.ServerConstants;
import org.wso2.wsas.admin.service.util.AppenderData;
import org.wso2.wsas.admin.service.util.LogData;
import org.wso2.wsas.admin.service.util.LoggerData;
import org.wso2.wsas.persistence.PersistenceManager;
import org.wso2.wsas.persistence.dataobject.AppenderDO;
import org.wso2.wsas.persistence.dataobject.LoggerDO;
import org.wso2.wsas.util.LoggingUtil;

import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 *
 */
public class LoggingAdmin {
    private static final Log log = LogFactory.getLog(LoggingAdmin.class);

    private PersistenceManager pm = new PersistenceManager();

    /**
     * Update the System Log Level
     *
     * @param logLevel
     * @param logPattern
     */
    public void updateSystemLog(String logLevel, String logPattern) throws AxisFault {

        if (log.isDebugEnabled()) {
            log.debug("Going to update log data with ==> Log level " + logLevel +
                      " log pattern " + logPattern);
        }

        Set appenderSet = new HashSet();

        // update root logger details
        Logger rootLogger = Logger.getRootLogger();
        rootLogger.setLevel(Level.toLevel(logLevel));
        pm.updateConfigurationProperty(ServerConstants.Logging.SYSTEM_LOG_LEVEL,
                                       logLevel);
        pm.updateConfigurationProperty(ServerConstants.Logging.SYSTEM_LOG_PATTERN,
                                       logPattern);
        addAppendersToSet(rootLogger.getAllAppenders(), appenderSet);

        // update logger and appender data, following are set
        // 1. log level of all the loggers to logLevel
        // 2. pattern of all the appenders to logpattern
        Enumeration loggers = LogManager.getLoggerRepository().getCurrentLoggers();
        Logger logger;
        LoggerDO loggerDO;

        while (loggers.hasMoreElements()) {
            logger = (Logger) loggers.nextElement();

            // we ignore all class level defined loggers
            if (logger.getLevel() != null) {
                addAppendersToSet(logger.getAllAppenders(), appenderSet);

                if ((logLevel != null) && (logLevel.trim().length() != 0)) {
                    loggerDO = pm.getLoggerDO(logger.getName());
                    logger.setLevel(Level.toLevel(logLevel));
                    pm.updateLoggerData(loggerDO.getName(), logLevel,
                                        loggerDO.getAdditivity());
                    if (log.isDebugEnabled()) {
                        log.debug("set the logger ==> " + logger.getName() +
                                  " to " + logger.getLevel().toString());
                    }
                }
            }
        }

        // update the appender data according to data stored in database
        Appender appender;
        AppenderDO appenderDO;

        for (Iterator iter = appenderSet.iterator(); iter.hasNext();) {
            appender = (Appender) iter.next();
            appenderDO = pm.getAppenderDO(appender.getName());

            if ((appender.getLayout() != null) &&
                (appender.getLayout() instanceof PatternLayout)) {
                if ((logPattern != null) && (logPattern.trim().length() != 0)) {
                    ((PatternLayout) appender.getLayout()).setConversionPattern(logPattern);
                    pm.updateAppenderData(appenderDO.getName(),
                                          logPattern,
                                          logLevel,
                                          appenderDO.getLogFileName(),
                                          appenderDO.getSysLogHost(),
                                          appenderDO.getFacility(),
                                          appenderDO.getIsFileAppender(),
                                          appenderDO.getIsSysLogAppender());
                }

                // set the threshold
                if (appender instanceof AppenderSkeleton) {
                    AppenderSkeleton appenderSkeleton = (AppenderSkeleton) appender;
                    appenderSkeleton.setThreshold(Level.toLevel(logLevel));
                    appenderSkeleton.activateOptions();
                }
            }
            if (log.isDebugEnabled()) {
                log.debug("set details for appender ==> " + appenderDO.toString());
            }
        }
    }

    /**
     * @param appenderName
     * @return the pattern string for a given appender
     */
    public AppenderData getAppenderData(String appenderName) {
        AppenderData appenderData = new AppenderData();
        AppenderDO appenderDO = pm.getAppenderDO(appenderName);
        appenderData.setName(appenderDO.getName());
        appenderData.setLogFile(appenderDO.getLogFileName());
        appenderData.setPattern(appenderDO.getPattern());
        appenderData.setIsFileAppender(appenderDO.getIsFileAppender());
        appenderData.setIsSysLogAppender(appenderDO.getIsSysLogAppender());
        appenderData.setFacility(appenderDO.getFacility());
        appenderData.setSysLogHost(appenderDO.getSysLogHost());
        appenderData.setThreshold(appenderDO.getThreshold());

        return appenderData;
    }

    public LoggerData getLoggerData(String loggerName) {
        LoggerData loggerData = new LoggerData();
        LoggerDO loggerDO = pm.getLoggerDO(loggerName);
        loggerData.setName(loggerDO.getName());
        loggerData.setLevel(loggerDO.getLogLevel());
        loggerData.setAditivity(loggerDO.getAdditivity());

        return loggerData;
    }

    /**
     * we receive all the parameters from the update appenders method but we
     * have to only update the relavent data.
     *
     * @param appenderName
     * @param appenderPattern
     * @param logFileName
     */
    public void updateAllAppenderData(String appenderName,
                                      String appenderPattern,
                                      String threshold,
                                      String logFileName,
                                      String sysLogHost,
                                      String facility) throws AxisFault {

        // get the existing appender and set the is File appender value
        AppenderDO appenderDO = pm.getAppenderDO(appenderName);

        if (appenderDO.getIsFileAppender()) {
            // Check if the file is valid
            logFileName = logFileName.replace('\\', '/');
            File logFile = new File(logFileName);
            if (!logFile.isAbsolute()) {
                if (logFileName.startsWith("./")) {
                    logFileName = logFileName.substring(2);
                }
                logFileName = (System.getProperty(ServerConstants.WSO2WSAS_HOME) + "/" +
                               logFileName).replace('\\', '/');
                logFile = new File(logFileName);
            }
            if (!logFile.exists()) {
                int lastIndex = logFileName.lastIndexOf("/");
                String msg = "Cannot create logfile " + logFileName +
                             ". Please verify that the log file name is valid " +
                             "and that you have read-write access to this file.";
                if (lastIndex != -1) {
                    String dirName = logFileName.substring(0, lastIndex);
                    File dir = new File(dirName);
                    if (!dir.exists() && !dir.mkdirs()) {
                        throw new AxisFault(msg);
                    }
                }
                try {
                    if (!logFile.createNewFile()) {
                        throw new AxisFault(msg);
                    }
                } catch (IOException e) {
                    throw new AxisFault(msg);
                }
            }
            pm.updateAppenderData(appenderName, appenderPattern, threshold,
                                  logFileName, null, null, true, false);
        } else if (appenderDO.getIsSysLogAppender()) {
            pm.updateAppenderData(appenderName, appenderPattern, threshold,
                                  null, sysLogHost, facility, false, true);
        } else {
            pm.updateAppenderData(appenderName, appenderPattern, threshold,
                                  null, null, null, false, false);
        }

        // update system appender data
        Set appenderSet = new HashSet();
        Logger rootLogger = Logger.getRootLogger();
        addAppendersToSet(rootLogger.getAllAppenders(), appenderSet);

        Enumeration loggers = LogManager.getCurrentLoggers();
        Logger logger;

        while (loggers.hasMoreElements()) {
            logger = (Logger) loggers.nextElement();
            if (logger.getLevel() != null) {
                addAppendersToSet(logger.getAllAppenders(), appenderSet);
            }
        }
        Appender appender = null;
        for (Iterator iter = appenderSet.iterator(); iter.hasNext();) {
            appender = (Appender) iter.next();

            if (appender.getName().equals(appenderName)) {
                break;
            }
        }

        if (appender != null) {
            if ((appender.getLayout() != null) &&
                (appender.getLayout() instanceof PatternLayout)) {
                ((PatternLayout) appender.getLayout()).setConversionPattern(appenderPattern);
            }

            if (appender instanceof FileAppender) {
                ((FileAppender) appender).setFile(logFileName);
                if (log.isDebugEnabled()) {
                    log.debug("change the logfile of the appender ==> " +
                              appender.getName() + " to " + logFileName);
                }
                ((FileAppender) appender).activateOptions();
            }

            if (appender instanceof SyslogAppender) {
                SyslogAppender syslogAppender = (SyslogAppender) appender;
                syslogAppender.setSyslogHost(sysLogHost);
                syslogAppender.setFacility(facility);
            }

            // set the threshold
            if (appender instanceof AppenderSkeleton) {
                AppenderSkeleton appenderSkeleton = (AppenderSkeleton) appender;
                appenderSkeleton.setThreshold(Level.toLevel(threshold));
                appenderSkeleton.activateOptions();
            }
        }
    }

    public void updateLoggerData(String loggerName,
                                 String loggerLevel,
                                 boolean additivity) {

        pm.updateLoggerData(loggerName, loggerLevel, additivity);

        //update logger data in current system
        Logger logger = LogManager.getLogger(loggerName);

        if (logger != null) {
            logger.setLevel(Level.toLevel(loggerLevel));
            logger.setAdditivity(additivity);
            if (log.isDebugEnabled()) {
                log.debug("Set the log level of logger ==>" + logger.getName() +
                          " to " + logger.getLevel().toString());
            }
        }
    }

    /**
     * The system log level
     *
     * @return The system log level
     * @throws AxisFault
     */
    public LogData getSystemLog() throws AxisFault {
        LogData logData = new LogData();

        // loading initial data
        logData.setLogLevel(pm.getConfigurationProperty(
                ServerConstants.Logging.SYSTEM_LOG_LEVEL));
        logData.setLogPattern(pm.getConfigurationProperty(
                ServerConstants.Logging.SYSTEM_LOG_PATTERN));

        // set the appenders
        AppenderData[] appenderData = getAllAppenderData();
        if (appenderData.length > 0) {
            logData.setAppenderData(appenderData);
            logData.setSelectedAppenderData(appenderData[0]);
        }

        // set the loggers
        LoggerData[] loggerData = getAllLoggerData();
        if (loggerData.length > 0) {
            logData.setLoggerData(loggerData);
            logData.setSelectedLoggerData(loggerData[0]);
        }

        return logData;
    }

    public void restoreDefaults() {
        LoggingUtil.restoreDefaults();
//        LoggingUtil.loadDefaultConfiguration();
    }

    private AppenderData[] getAllAppenderData() {
        List appenderDOs = pm.getAllAppenders();
        AppenderData[] appenderDataArray = new AppenderData[appenderDOs.size()];
        AppenderDO appeanderDO;

        for (int i = 0; i < appenderDOs.size(); i++) {
            appeanderDO = (AppenderDO) appenderDOs.get(i);
            appenderDataArray[i] = new AppenderData();
            appenderDataArray[i].setName(appeanderDO.getName());
            appenderDataArray[i].setPattern(appeanderDO.getPattern());
            appenderDataArray[i].setFacility(appeanderDO.getFacility());
            appenderDataArray[i].setLogFile(appeanderDO.getLogFileName());
            appenderDataArray[i].setIsFileAppender(appeanderDO.getIsFileAppender());
            appenderDataArray[i].setIsSysLogAppender(appeanderDO.getIsSysLogAppender());
            appenderDataArray[i].setSysLogHost(appeanderDO.getSysLogHost());
            appenderDataArray[i].setThreshold(appeanderDO.getThreshold());
        }

        return appenderDataArray;
    }

    private LoggerData[] getAllLoggerData() {
        List loggerDOs = pm.getAllLoggers();
        LoggerData[] loggerDatas = new LoggerData[loggerDOs.size()];
        LoggerDO loggerDO;

        for (int i = 0; i < loggerDOs.size(); i++) {
            loggerDO = (LoggerDO) loggerDOs.get(i);
            loggerDatas[i] = new LoggerData(loggerDO.getName(),
                                            loggerDO.getLogLevel(),
                                            loggerDO.getAdditivity());
        }

        return loggerDatas;
    }

    private void addAppendersToSet(Enumeration appenders, Set appenderSet) {
        Appender appender;
        while (appenders.hasMoreElements()) {
            appender = (Appender) appenders.nextElement();
            appenderSet.add(appender);
            if (log.isDebugEnabled()) {
                log.debug("Add appender ==> " + appender.getName() + " to appender set");
            }
        }
    }
}
